#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>

#include "cJSON.h"
#include "semaphore.h"

#include "aiot_state_api.h"
#include "aiot_sysdep_api.h"
#include "aiot_mqtt_api.h"
#include "aiot_dm_api.h"

#include "log_util.h"
#include "linkface_sdk.h"

#ifdef __APPLE__
static sem_t *g_sem_linkkit_loop;
#else
static sem_t g_sem_linkkit_loop;
#endif
static int g_flag_is_running = 0;

static void *g_mqtt_handle = NULL;
static void *g_dm_handle = NULL;
static pthread_t g_mqtt_process_thread;
static pthread_t g_mqtt_recv_thread;
static uint8_t g_mqtt_process_thread_running = 0;
static uint8_t g_mqtt_recv_thread_running = 0;

static const char *ali_ca_cert = \
{
        \
    "-----BEGIN CERTIFICATE-----\r\n"
    "MIIDdTCCAl2gAwIBAgILBAAAAAABFUtaw5QwDQYJKoZIhvcNAQEFBQAwVzELMAkG\r\n" \
    "A1UEBhMCQkUxGTAXBgNVBAoTEEdsb2JhbFNpZ24gbnYtc2ExEDAOBgNVBAsTB1Jv\r\n" \
    "b3QgQ0ExGzAZBgNVBAMTEkdsb2JhbFNpZ24gUm9vdCBDQTAeFw05ODA5MDExMjAw\r\n" \
    "MDBaFw0yODAxMjgxMjAwMDBaMFcxCzAJBgNVBAYTAkJFMRkwFwYDVQQKExBHbG9i\r\n" \
    "YWxTaWduIG52LXNhMRAwDgYDVQQLEwdSb290IENBMRswGQYDVQQDExJHbG9iYWxT\r\n" \
    "aWduIFJvb3QgQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDaDuaZ\r\n" \
    "jc6j40+Kfvvxi4Mla+pIH/EqsLmVEQS98GPR4mdmzxzdzxtIK+6NiY6arymAZavp\r\n" \
    "xy0Sy6scTHAHoT0KMM0VjU/43dSMUBUc71DuxC73/OlS8pF94G3VNTCOXkNz8kHp\r\n" \
    "1Wrjsok6Vjk4bwY8iGlbKk3Fp1S4bInMm/k8yuX9ifUSPJJ4ltbcdG6TRGHRjcdG\r\n" \
    "snUOhugZitVtbNV4FpWi6cgKOOvyJBNPc1STE4U6G7weNLWLBYy5d4ux2x8gkasJ\r\n" \
    "U26Qzns3dLlwR5EiUWMWea6xrkEmCMgZK9FGqkjWZCrXgzT/LCrBbBlDSgeF59N8\r\n" \
    "9iFo7+ryUp9/k5DPAgMBAAGjQjBAMA4GA1UdDwEB/wQEAwIBBjAPBgNVHRMBAf8E\r\n" \
    "BTADAQH/MB0GA1UdDgQWBBRge2YaRQ2XyolQL30EzTSo//z9SzANBgkqhkiG9w0B\r\n" \
    "AQUFAAOCAQEA1nPnfE920I2/7LqivjTFKDK1fPxsnCwrvQmeU79rXqoRSLblCKOz\r\n" \
    "yj1hTdNGCbM+w6DjY1Ub8rrvrTnhQ7k4o+YviiY776BQVvnGCv04zcQLcFGUl5gE\r\n" \
    "38NflNUVyRRBnMRddWQVDf9VMOyGj/8N7yy5Y0b2qvzfvGn9LhJIZJrglfCm7ymP\r\n" \
    "AbEVtQwdpf5pLGkkeB6zpxxxYu7KyJesF12KwvhHhm4qxFYxldBniYUr+WymXUad\r\n" \
    "DKqC5JlR3XC321Y9YeRq4VzW9v493kHMB65jUr9TU/Qr6cf9tveCX4XSQRjbgbME\r\n" \
    "HMUfpIBvFSDJ3gyICh3WZlXi/EjJKSZp4A==\r\n" \
    "-----END CERTIFICATE-----"
};

extern aiot_sysdep_portfile_t g_aiot_sysdep_portfile;

extern face_event_handler(const size_t service_id_len, const char *service_id, const size_t request_len,
                          const char *request, char **response, size_t *response_len);

/* 日志回调函数, SDK的日志会从这里输出 */
int32_t linksdk_state_logcb_adapter(int32_t code, char *message) {
    LOG_INFO("linksdk adapter log info code: %d,message: %s\n", code, message);
    return 0;
}

/* 执行aiot_mqtt_process的线程, 包含心跳发送和QoS1消息重发 */
void *mqtt_process_thread_adapter(void *args) {
    int32_t res = STATE_SUCCESS;

    while (g_mqtt_process_thread_running) {
        res = aiot_mqtt_process(args);
        if (res == STATE_USER_INPUT_EXEC_DISABLED) {
            break;
        }
        sleep(1);
    }
    return NULL;
}

/* 执行aiot_mqtt_recv的线程, 包含网络自动重连和从服务器收取MQTT消息 */
void *mqtt_recv_thread_adapter(void *args) {
    int32_t res = STATE_SUCCESS;

    while (g_mqtt_recv_thread_running) {
        res = aiot_mqtt_recv(args);
        if (res < STATE_SUCCESS) {
            if (res == STATE_USER_INPUT_EXEC_DISABLED) {
                break;
            }
            sleep(1);
        }
    }
    return NULL;
}

/* 用户数据接收处理回调函数 */
static void dm_recv_handler_adapter(void *dm_handle, const aiot_dm_recv_t *recv, void *userdata) {
    LOG_INFO("dm_recv_handler_adapter, type = %d\r\n", recv->type);

    switch (recv->type) {
        /* 属性上报, 事件上报, 获取期望属性值或者删除期望属性值的应答 */
        case AIOT_DMRECV_GENERIC_REPLY: {
            LOG_INFO("dm_recv_handler_adapter msg_id = %d, code = %d, data = %.*s, message = %.*s\r\n",
                     recv->data.generic_reply.msg_id,
                     recv->data.generic_reply.code,
                     recv->data.generic_reply.data_len,
                     recv->data.generic_reply.data,
                     recv->data.generic_reply.message_len,
                     recv->data.generic_reply.message);
        }
            break;

            /* 属性设置 */
        case AIOT_DMRECV_PROPERTY_SET: {
            LOG_INFO("dm_recv_handler_adapter msg_id = %ld, params = %.*s\r\n",
                     (unsigned long) recv->data.property_set.msg_id,
                     recv->data.property_set.params_len,
                     recv->data.property_set.params);
        }
            break;

            /* 异步服务调用 */
        case AIOT_DMRECV_ASYNC_SERVICE_INVOKE: {
            LOG_INFO("dm_recv_handler_adapter msg_id = %ld, service_id = %s, params = %.*s\r\n",
                     (unsigned long) recv->data.async_service_invoke.msg_id,
                     recv->data.async_service_invoke.service_id,
                     recv->data.async_service_invoke.params_len,
                     recv->data.async_service_invoke.params);
        }
            break;

            /* 同步服务调用 */
        case AIOT_DMRECV_SYNC_SERVICE_INVOKE: {
            LOG_INFO("dm_recv_handler_adapter msg_id = %ld, rrpc_id = %s, service_id = %s, params = %.*s\r\n",
                     (unsigned long) recv->data.sync_service_invoke.msg_id,
                     recv->data.sync_service_invoke.rrpc_id,
                     recv->data.sync_service_invoke.service_id,
                     recv->data.sync_service_invoke.params_len,
                     recv->data.sync_service_invoke.params);
            {
                if (g_flag_is_running == 0) {
                    LOG_ERROR("dm_recv_handler_adapter please init link sdk first.\r\n");
                    return;
                }
                char *service_id = recv->data.sync_service_invoke.service_id;
                size_t service_id_len = strlen(service_id);
                char *request = recv->data.sync_service_invoke.params;
                size_t request_len = strlen(request);
                char *response = NULL;
                size_t response_len = 0;
                int handler_res = face_event_handler(service_id_len, service_id, request_len, request, &response,
                                                     &response_len);
                if (handler_res != 0) {
                    LOG_ERROR("dm_recv_handler_adapter face_event_handler_adapter failed handler_res: %d\r\n",
                              handler_res);
                    return;
                }

                aiot_dm_msg_t msg;
                memset(&msg, 0, sizeof(aiot_dm_msg_t));
                msg.type = AIOT_DMMSG_SYNC_SERVICE_REPLY;
                msg.data.sync_service_reply.rrpc_id = recv->data.sync_service_invoke.rrpc_id;
                msg.data.sync_service_reply.msg_id = recv->data.sync_service_invoke.msg_id;
                msg.data.sync_service_reply.code = 200;
                msg.data.sync_service_reply.service_id = service_id;
                msg.data.sync_service_reply.data = response;
                int32_t res = aiot_dm_send(g_dm_handle, &msg);
                if (res < 0) {
                    LOG_ERROR("dm_recv_handler_adapter aiot_dm_send failed res: -0x%04X\r\n", -res);
                }
                free(response);
            }
        }
            break;

            /* 下行二进制数据 */
        case AIOT_DMRECV_RAW_DATA: {
            LOG_INFO("dm_recv_handler_adapter raw data len = %d\r\n", recv->data.raw_data.data_len);
        }
            break;

            /* 二进制格式的同步服务调用, 比单纯的二进制数据消息多了个rrpc_id */
        case AIOT_DMRECV_RAW_SYNC_SERVICE_INVOKE: {
            LOG_INFO("dm_recv_handler_adapter raw sync service rrpc_id = %s, data_len = %d\r\n",
                     recv->data.raw_service_invoke.rrpc_id,
                     recv->data.raw_service_invoke.data_len);
        }
            break;

            /* 上行二进制数据后, 云端的回复报文 */
        case AIOT_DMRECV_RAW_DATA_REPLY: {
            LOG_INFO("dm_recv_handler_adapter receive reply for up_raw msg, data len = %d\r\n",
                     recv->data.raw_data.data_len);
        }
            break;

        default:
            break;
    }
}

/* MQTT事件回调函数, 当网络连接/重连/断开时被触发, 事件定义见core/aiot_mqtt_api.h */
void mqtt_event_handler_adapter(void *handle, const aiot_mqtt_event_t *event, void *userdata) {
    switch (event->type) {
        /* SDK因为用户调用了aiot_mqtt_connect()接口, 与mqtt服务器建立连接已成功 */
        case AIOT_MQTTEVT_CONNECT: {
            LOG_INFO("mqtt_event_handler_adapter AIOT_MQTTEVT_CONNECT\n");
        }
            break;

            /* SDK因为网络状况被动断连后, 自动发起重连已成功 */
        case AIOT_MQTTEVT_RECONNECT: {
            LOG_INFO("mqtt_event_handler_adapter AIOT_MQTTEVT_RECONNECT\n");
        }
            break;

            /* SDK因为网络的状况而被动断开了连接, network是底层读写失败, heartbeat是没有按预期得到服务端心跳应答 */
        case AIOT_MQTTEVT_DISCONNECT: {
            char *cause = (event->data.disconnect == AIOT_MQTTDISCONNEVT_NETWORK_DISCONNECT) ? ("network disconnect")
                                                                                             : ("heartbeat disconnect");
            LOG_INFO("mqtt_event_handler_adapter AIOT_MQTTEVT_DISCONNECT: %s\n", cause);
        }
            break;

        default: {
            LOG_ERROR("mqtt_event_handler_adapter event type=%d not exists\n", event->type);
        }
    }
}

void *mqtt_loop_adapter() {
#ifdef __APPLE__
    sem_post(g_sem_linkkit_loop);
#else
    sem_post(&g_sem_linkkit_loop);
#endif

    while (1 == g_flag_is_running) {
        sleep(1);
    }
    LOG_INFO("mqtt_loop_adapter mqtt client exit.\n");

#ifdef __APPLE__
    sem_post(g_sem_linkkit_loop);
#else
    sem_post(&g_sem_linkkit_loop);
#endif
    return NULL;
}

int linksdk_init_adapter(struct iotx_lf_config_t *cfg) {
    int32_t res = STATE_SUCCESS;
    char host[100] = {0}; /* 用这个数组拼接设备连接的云平台站点全地址, 规则是 ${productKey}.iot-as-mqtt.cn-shanghai.aliyuncs.com */
    uint16_t port = 443;      /* 无论设备是否使用TLS连接阿里云平台, 目的端口都是443 */
    aiot_sysdep_network_cred_t cred; /* 安全凭据结构体, 如果要用TLS, 这个结构体中配置CA证书等参数 */
    uint8_t post_reply = 1;

    /* 配置SDK的底层依赖 */
    aiot_sysdep_set_portfile(&g_aiot_sysdep_portfile);
    /* 配置SDK的日志输出 */
    aiot_state_set_logcb(linksdk_state_logcb_adapter);

    /* 创建SDK的安全凭据, 用于建立TLS连接 */
    memset(&cred, 0, sizeof(aiot_sysdep_network_cred_t));
    cred.option = AIOT_SYSDEP_NETWORK_CRED_SVRCERT_CA;  /* 使用RSA证书校验MQTT服务端 */
    cred.max_tls_fragment = 16384; /* 最大的分片长度为16K, 其它可选值还有4K, 2K, 1K, 0.5K */
    cred.sni_enabled = 1;                               /* TLS建连时, 支持Server Name Indicator */
    cred.x509_server_cert = ali_ca_cert;                 /* 用来验证MQTT服务端的RSA根证书 */
    cred.x509_server_cert_len = strlen(ali_ca_cert);     /* 用来验证MQTT服务端的RSA根证书长度 */

    /* 创建1个MQTT客户端实例并内部初始化默认参数 */
    g_mqtt_handle = aiot_mqtt_init();
    if (g_mqtt_handle == NULL) {
        LOG_ERROR("linksdk_init_adapter aiot_mqtt_init failed\r\n");
        return -1;
    }

    mqtt_host(cfg, host);
    /* 配置MQTT服务器地址 */
    aiot_mqtt_setopt(g_mqtt_handle, AIOT_MQTTOPT_HOST, (void *) host);
    /* 配置MQTT服务器端口 */
    aiot_mqtt_setopt(g_mqtt_handle, AIOT_MQTTOPT_PORT, (void *) &port);
    /* 配置设备productKey */
    aiot_mqtt_setopt(g_mqtt_handle, AIOT_MQTTOPT_PRODUCT_KEY, (void *) cfg->product_key);
    /* 配置设备deviceName */
    aiot_mqtt_setopt(g_mqtt_handle, AIOT_MQTTOPT_DEVICE_NAME, (void *) cfg->device_name);
    /* 配置设备deviceSecret */
    aiot_mqtt_setopt(g_mqtt_handle, AIOT_MQTTOPT_DEVICE_SECRET, (void *) cfg->device_secret);
    /* 配置网络连接的安全凭据, 上面已经创建好了 */
    aiot_mqtt_setopt(g_mqtt_handle, AIOT_MQTTOPT_NETWORK_CRED, (void *) &cred);
    /* 配置MQTT事件回调函数 */
    aiot_mqtt_setopt(g_mqtt_handle, AIOT_MQTTOPT_EVENT_HANDLER, (void *) mqtt_event_handler_adapter);

    /* 创建DATA-MODEL实例 */
    g_dm_handle = aiot_dm_init();
    if (g_dm_handle == NULL) {
        LOG_ERROR("linksdk_init_adapter aiot_dm_init failed\r\n");
        return -1;
    }
    /* 配置MQTT实例句柄 */
    aiot_dm_setopt(g_dm_handle, AIOT_DMOPT_MQTT_HANDLE, g_mqtt_handle);
    /* 配置消息接收处理回调函数 */
    aiot_dm_setopt(g_dm_handle, AIOT_DMOPT_RECV_HANDLER, (void *) dm_recv_handler_adapter);

    /* 配置是云端否需要回复post_reply给设备. 如果为1, 表示需要云端回复, 否则表示不回复 */
    aiot_dm_setopt(g_dm_handle, AIOT_DMOPT_POST_REPLY, (void *) &post_reply);

    /* 与服务器建立MQTT连接 */
    res = aiot_mqtt_connect(g_mqtt_handle);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_dm_deinit(&g_dm_handle);
        aiot_mqtt_deinit(&g_mqtt_handle);
        LOG_ERROR("linksdk_init_adapter aiot_mqtt_connect failed: -0x%04X\r\n", -res);
        return -1;
    }

    /* 创建一个单独的线程, 专用于执行aiot_mqtt_process, 它会自动发送心跳保活, 以及重发QoS1的未应答报文 */
    g_mqtt_process_thread_running = 1;
    res = pthread_create(&g_mqtt_process_thread, NULL, mqtt_process_thread_adapter, g_mqtt_handle);
    if (res < 0) {
        LOG_ERROR("linksdk_init_adapter pthread_create mqtt_process_thread_adapter failed: %d\r\n", res);
        aiot_dm_deinit(&g_dm_handle);
        aiot_mqtt_disconnect(g_mqtt_handle);
        aiot_mqtt_deinit(&g_mqtt_handle);
        return -1;
    }

    /* 创建一个单独的线程用于执行aiot_mqtt_recv, 它会循环收取服务器下发的MQTT消息, 并在断线时自动重连 */
    g_mqtt_recv_thread_running = 1;
    res = pthread_create(&g_mqtt_recv_thread, NULL, mqtt_recv_thread_adapter, g_mqtt_handle);
    if (res < 0) {
        LOG_ERROR("linksdk_init_adapter pthread_create mqtt_recv_thread_adapter failed: %d\r\n", res);
        aiot_dm_deinit(&g_dm_handle);
        aiot_mqtt_disconnect(g_mqtt_handle);
        aiot_mqtt_deinit(&g_mqtt_handle);
        return -1;
    }

    // 外部初始化LinkSDK，需要把mqtt_handle赋值给Linkface SDK
    cfg->mqtt_handle = g_mqtt_handle;
    // 外部初始化LinkSDK，需要把dm_handle赋值给Linkface SDK
    cfg->dm_handle = g_dm_handle;
    return res;
}

int linkkit_init_adapter(struct iotx_lf_config_t *cfg) {
    if (linksdk_init_adapter(cfg) < STATE_SUCCESS) {
        LOG_ERROR("LinkSDK init failed.\r\n");
        return -1;
    }

#ifdef __APPLE__
    g_sem_linkkit_loop = sem_open("sem_linkkit_loop", O_CREAT);
#else
    sem_init(&g_sem_linkkit_loop, 0, 0);
#endif
    g_flag_is_running = 1;
    pthread_t thread_id = 0;
    pthread_create(&thread_id, NULL, mqtt_loop_adapter, NULL);
    pthread_detach(thread_id);
    if (thread_id > 0) {
        //必须等到linkkit_loop线程起来之后才返回。
#ifdef __APPLE__
        sem_wait(g_sem_linkkit_loop);
#else
        sem_wait(&g_sem_linkkit_loop);
#endif
        return 1;
    } else {

#ifdef __APPLE__
        sem_close(g_sem_linkkit_loop);
#else
        sem_destroy(&g_sem_linkkit_loop);
#endif
        linkkit_destroy_adapter();
        return -1;
    }
}

int linkkit_destroy_adapter() {
    /* 停止收发动作 */
    g_mqtt_process_thread_running = 0;
    g_mqtt_recv_thread_running = 0;

    /* 断开MQTT连接, 一般不会运行到这里 */
    int32_t res = aiot_mqtt_disconnect(g_mqtt_handle);
    if (res < STATE_SUCCESS) {
        aiot_dm_deinit(&g_dm_handle);
        aiot_mqtt_deinit(&g_mqtt_handle);
        LOG_ERROR("linkkit_destroy_adapter aiot_mqtt_disconnect failed: -0x%04X\r\n", -res);
        return -1;
    }

    /* 销毁DATA-MODEL实例, 一般不会运行到这里 */
    res = aiot_dm_deinit(&g_dm_handle);
    if (res < STATE_SUCCESS) {
        LOG_ERROR("linkkit_destroy_adapter aiot_dm_deinit failed: -0x%04X\r\n", -res);
        return -1;
    }

    /* 销毁MQTT实例, 一般不会运行到这里 */
    res = aiot_mqtt_deinit(&g_mqtt_handle);
    if (res < STATE_SUCCESS) {
        LOG_ERROR("linkkit_destroy_adapter aiot_mqtt_deinit failed: -0x%04X\r\n", -res);
        return -1;
    }

    pthread_join(g_mqtt_process_thread, NULL);
    pthread_join(g_mqtt_recv_thread, NULL);
    return 0;
}

